import io
import numpy as np
import torch 
torch.manual_seed(42)
torch.set_num_threads(1)
from torch.utils.checkpoint import checkpoint

from utils import gauge, trace

def build_tensor(K, H):
    c = torch.sqrt(torch.cosh(K))
    s = torch.sqrt(torch.sinh(K))
    Q = [c, s, c, -s]
    Q = torch.stack(Q).view(2, 2)

    M = [torch.exp(H), torch.exp(-H)]
    M = torch.stack(M).view(2)

    T = torch.einsum('a,ai,aj,ak,al->ijkl', (M, Q, Q, Q, Q))
    return T

def renormalize(*tensors):
    T, chi = tensors

    torch.set_grad_enabled(args.enable_grad)
    vl, el = gauge(T, chi,'l')
    vr, er = gauge(T, chi,'r')
    U = vl if (el < er) else vr
    torch.set_grad_enabled(True)

    T = torch.einsum('axob,amz,moyn,bnw->xwzy',(T, U, T, U))

    #TU1 = torch.tensordot(T, U, dims=([0], [0]))
    #TU2 = torch.tensordot(T, U, dims=([3], [1]))
    #T = torch.tensordot(TU1, TU2, dims=([1,2,3], [1,3,0]))
    #T = T.permute(0,3,1,2).contiguous()
    return T

class Ising(torch.nn.Module):
    def __init__(self, chi, niter, dtype=torch.float64, device='cpu', use_checkpoint=False):
        super(Ising, self).__init__()
    
        self.D = 2
        self.chi = chi # cutoff 
        self.niter = niter
        self.dtype = dtype
        self.device = device

    def forward(self, T, use_checkpoint=False):
        '''
        '''
        lnZ = 0.0
        for step in range(self.niter):
            f = T.abs().max()
            lnZ += 2.0**(-step)*torch.log(f)
            T = T / f
           
            tensors = T, torch.tensor(self.chi)
            if use_checkpoint: # use checkpoint to save memory
                T = checkpoint(renormalize, *tensors)
            else:
                T = renormalize(*tensors)

        lnZ += torch.log(trace(T))/2.0**args.niter
        return lnZ


def main(beta, h=0.):

    beta = torch.tensor(beta, dtype=dtype, device=device)
    h = torch.tensor(h, dtype=dtype, device=device).requires_grad_()

    model = Ising(args.chi, args.niter, dtype=dtype, device=device)

    T = build_tensor(beta, beta*h)
    lnZ = model.forward(T, args.use_checkpoint)
    lnZ.backward(create_graph=True)
    m = (h.grad).item() # dlnZ /d h  
    
    h.grad.backward()
    m2 = (h.grad).item() -m # second order gradient d^2 lnZ/dh^2 
    
    m = torch.sqrt(m2/beta**2/2.0**args.niter)
    print ('beta, lnZ, m', beta.item(), lnZ.item(), m.item())

    return lnZ, m

def scan_beta():

    key = 'D' + str(args.chi) + \
          'N' + str(args.niter) 

    if args.enable_grad:
        key += '_disometry'

    with io.open(key+'.log', 'a', buffering=1, newline='\n') as logfile:
        for beta in np.linspace(0.44, 0.48, 51):
            lnZ, m = main(beta)
            logfile.write((3*'{:.15f} ').format(beta, lnZ.item(), m.item()) + u'\n')

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    
    parser.add_argument("-beta", type=float, default=None, help="beta")
    parser.add_argument("-chi", type=int, default=20, help="chi")
    parser.add_argument("-niter", type=int, default=80, help="niter")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-use_checkpoint", action='store_true', help="use checkpoint")
    parser.add_argument("-enable_grad", action='store_true', help="diff into isometry")
    parser.add_argument("-cuda", type=int, default=-1, help="GPU #")
    args = parser.parse_args()

    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    print ('use', dtype)
    print ('use checkpoint', args.use_checkpoint)
   
    if args.beta is not None:
        main(args.beta)
    else:
        scan_beta()
