import numpy as np
import torch 
torch.manual_seed(42)

def gauge(T, Dcut, side):
    D = T.shape[0]
    M = torch.einsum('ludr,adcb->laucrb', (T, T)).contiguous()
    if side=='l':
        M = M.view(D**2, -1)
        M = M@M.t()
    elif side=='r':
        M = M.view(-1, D**2)
        M = M.t()@M

    U, S, V = torch.svd(M)
    Dnew = min(D**2, Dcut)
    truncation = S[Dnew:].sum() / S.sum()
    return U[:, :Dnew].view(D, D, Dnew), truncation

class Ising(torch.nn.Module):
    def __init__(self, chi, niter, dtype=torch.float64, device='cpu', use_checkpoint=False):
        super(Ising, self).__init__()
    
        self.D = 2
        self.chi = chi
        self.niter = niter
        self.dtype = dtype
        self.device = device

        B = 0.01*torch.randn(niter, self.chi, self.chi, self.chi, dtype=dtype, device=device)
        self.A = torch.nn.Parameter(B)

    def isometry(self, step):
        chi = min(self.D**(2**step), self.chi) 
        chi_new = min(self.chi, chi**2)
        return self.A[step, :chi, :chi, :chi_new]

    def set_isometry(self, step, U):
        chi = min(self.D**(2**step), self.chi) 
        chi_new = U.shape[-1]
        self.A[step, :chi, :chi, :chi_new] = U

    def init_isometry(self, T):
        lnZ = 0.0
        for step in range(self.niter):
            f = T.abs().max()
            lnZ += 2**(-step)*torch.log(f)
            T = T / f

            #contract up-down
            vl, el = gauge(T, self.chi,'l')
            vr, er = gauge(T, self.chi,'r')

            U = vl if (el < er) else vr
            self.set_isometry(step, U)
            T = torch.einsum('axob,amz,moyn,bnw->zxyw',(T, U, T, U))
           
        #build 1D transfer matrix 
        A = torch.zeros(T.shape[0], T.shape[3], dtype=T.dtype, device=T.device)
        for y in range(T.shape[1]):
            A = A+ T[:, y, y, :]
        A = (A + A.t())/2 

        lamb_up, _ = torch.symeig(A, eigenvectors=True)
        lnZ = lnZ + torch.log(lamb_up[-1])/2**self.niter
        return lnZ
        
    def forward(self, T):
        
        W = torch.eye(self.D, self.D, device=self.device, dtype=self.dtype)

        lnZ = 0.0
        for n in range(self.niter):
            f = T.abs().max()
            T = T/f 
            lnZ += 2**(-n)*torch.log(f)

            f = W.abs().max()
            W = W/f
            lnZ -= 2**(-n)*torch.log(f)

            U = self.isometry(n) # isometry for this step

            T = torch.einsum('axob,amz,moyn,bnw->zxyw',(T, U, T, U))
            W = torch.einsum('ab,amz,mn,bnw->zw',(W, U, W, U))
       
        #build 1D transfer matrix 
        A = torch.zeros(T.shape[0], T.shape[3], dtype=T.dtype, device=T.device)
        for y in range(T.shape[1]):
            A = A+ T[:, y, y, :]
        A = (A + A.t())/2 

        lamb_up, _ = torch.symeig(A, eigenvectors=True)
        lamb_dn, _ = torch.symeig(W, eigenvectors=True)
        lnZ = lnZ + torch.log(lamb_up[-1]/lamb_dn[-1])/2**self.niter
        return lnZ

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    
    parser.add_argument("-beta", type=float, default=0.44, help="beta")
    parser.add_argument("-chi", type=int, default=20, help="chi")
    parser.add_argument("-niter", type=int, default=20, help="niter")
    parser.add_argument("-Nepochs", type=int, default=100, help="Nepochs")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="GPU #")
    args = parser.parse_args()

    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    print ('use', dtype)

    model = Ising(args.chi, args.niter, dtype=dtype, device=device)

    optimizer = torch.optim.LBFGS(model.parameters(), max_iter=20, line_search_fn='strong_wolfe')
    params = list(model.parameters())
    params = list(filter(lambda p: p.requires_grad, params))
    nparams = sum([np.prod(p.size()) for p in params])
    print ('total nubmer of trainable parameters:', nparams)

    # initial tensor
    K = torch.tensor([args.beta], dtype=dtype, device=device)
    c = torch.sqrt(torch.cosh(K)/2.)
    s = torch.sqrt(torch.sinh(K)/2.)
    M = torch.stack([torch.cat([c+s, c-s]), torch.cat([c-s, c+s])])
    T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M)) 
    
    with torch.no_grad():
        lnZ = model.init_isometry(T)
        print ('initial loss:', -lnZ.item())

    def closure():
        optimizer.zero_grad()
        loss = -model.forward(T) # loss = -lnZ
        loss.backward()
        return loss

    for epoch in range(args.Nepochs):
        loss = optimizer.step(closure)
        print (epoch, loss.item())
